from ..MetaDataObject import MetaDataObject
from ..MetaDataObject.versions.Template801 import Template801
from ..MetaDataObject.versions.Template802 import Template802
from ..MetaDataObject.versions.Template803 import Template803


class Template(MetaDataObject):
    versions = {
        '801': Template801,
        '802': Template802,
        '803': Template803
    }



import os
from base64 import b64decode, b64encode

from .. import helper
from ..MetaDataObject.core.Simple import Simple
from ..ext_exception import ExtException


class CommonPicture(Simple):
    def __init__(self):
        super(Simple, self).__init__()
        self.data = None
        self.raw_data = None

    def decode_object(self, src_dir, file_name, dest_dir, dest_path, version, header_data):
        try:
            super(CommonPicture, self).decode_object(src_dir, file_name, dest_dir, dest_path, version, header_data)
            try:
                self.raw_data = helper.json_read(src_dir, f'{self.header["uuid"]}.0.json')
            except FileNotFoundError:
                return
            if self.raw_data[0][2] and self.raw_data[0][2][0][0]:
                self.data = b64decode(self.raw_data[0][2][0][0][8:])
                extension = helper.get_extension_from_comment(self.header['comment'])
                _dest_dir = os.path.join(dest_dir, dest_path)
                if dest_dir:
                    helper.bin_write(self.data, _dest_dir, f'{self.header["name"]}.{extension}')
        except Exception as err:
            raise ExtException(parent=err)

    def encode_object(self, src_dir, file_name, dest_dir, version):
        super(CommonPicture, self).encode_object(src_dir, file_name, dest_dir, version)

        extension = helper.get_extension_from_comment(self.header['comment'])
        try:
            bin_data = helper.bin_read(src_dir, f'{self.header["name"]}.{extension}')
            self.raw_data = [
                [
                    "1",
                    [
                        "0",
                        "0",
                        "-1",
                        "-1"
                    ],
                    [
                        [
                            "#base64:" + b64encode(bin_data).decode(encoding='utf-8')
                        ]
                    ]
                ]
            ]
            helper.json_write(self.raw_data, dest_dir, f'{self.header["uuid"]}.0.json')
        except FileNotFoundError:
            pass



from ..MetaDataObject.core.Simple import Simple


class ChartOfCharacteristicType(Simple):
    ext_code = {
        'mgr': '16',  # модуль менеджера Плана видов характеристик
        'obj': '15',  # модуль менеджера Плана видов характеристик
        'pre': '7',  # предопределенные Плана видов характеристик
    }

    @classmethod
    def get_decode_header(cls, header_data):
        return header_data[0][1][13][1]



import os

from . import helper
from .MetaObject.Configuration803 import Configuration803
from .MetaObject.ConfigurationExtension803 import ConfigurationExtension803
from .MetaObject.ExternalDataProcessor801 import ExternalDataProcessor801
from .MetaObject.ExternalDataProcessor802 import ExternalDataProcessor802
from .MetaObject.ExternalDataProcessor803 import ExternalDataProcessor803
from .ext_exception import ExtException
from .metadata_types import MetaDataTypes

available_types = {
    'ExternalDataProcessor801': ExternalDataProcessor801,
    'ExternalDataProcessor802': ExternalDataProcessor802,
    'ExternalDataProcessor803': ExternalDataProcessor803,
    'Configuration803': Configuration803
}


class Decoder:
    @staticmethod
    def detect_version(src_dir):
        if os.path.isfile(os.path.join(src_dir, 'root.json')):
            version = helper.json_read(src_dir, 'version.json')
            if version[0][0][0] == "216":
                _tmp = len(version[0][0])
                root = helper.json_read(src_dir, 'root.json')
                header = helper.json_read(src_dir, f'{root[0][1]}.json')
                obj_type = MetaDataTypes(header[0][3][0])
                if _tmp == 2:
                    obj_version = '802'
                elif _tmp == 3:
                    obj_version = version[0][0][2][0][:3]
                    if len(obj_version) == 3:
                        pass
                    elif obj_version == '2':
                        obj_version = '802'
                    else:
                        raise Exception(f'Not supported version {obj_version}')
                else:
                    raise Exception(f'Not supported version {_tmp}')
                try:
                    return available_types[f'{obj_type.name}{obj_version}']
                except KeyError:
                    raise Exception(f'Not supported version {obj_type.name}{obj_version}')
            elif version[0][0][0] == "106":
                return ExternalDataProcessor801()
        if os.path.isfile(os.path.join(src_dir, 'configinfo.json')):
            version = helper.json_read(src_dir, 'configinfo.json')
            if version[0][1][0] == "216":
                if len(version[0][1]) == 3:
                    obj_version = version[0][1][2][0]
                    if obj_version[:3] == '803':
                        return ConfigurationExtension803()
        raise Exception('Не удалось определить парсер')

    @classmethod
    def decode(cls, src_dir, dest_dir, *, pool=None, version=None):
        decoder = cls.detect_version(src_dir)
        helper.clear_dir(dest_dir)
        tasks = decoder.decode(src_dir, dest_dir,
                               version=version)  # возвращает список вложенных объектов MetaDataObject
        while tasks:  # многопоточно рекурсивно декодируем вложенные объекты MetaDataObject
            tasks_list = helper.run_in_pool(cls.decode_include, tasks, pool)
            tasks = helper.list_merge(*tasks_list)

    @classmethod
    def decode_include(cls, include_type, decode_params):
        try:
            handler = helper.get_class_metadata_object(include_type)
            handler = handler.get_version(decode_params[4])
            tasks = handler.decode(*decode_params)
            return tasks
        except ExtException as err:
            raise ExtException(
                parent=err,
                message=f'{err.message} - {err.detail}',
                detail=f'decode include {include_type}',
                dump={'decode_params': decode_params},
                action=f'{cls.__name__}.decode_include'
            )
        except Exception as err:
            raise ExtException(parent=err, action=f'{cls.__name__}.decode_include')

    @classmethod
    def encode(cls, src_dir, dest_dir, *, pool=None, version=None, file_name=None, gui=None, **kwargs):
        helper.clear_dir(dest_dir)
        encoder = cls.get_encoder(src_dir, '803' if version is None else version[:3])
        # возвращает список вложенных объектов MetaDataObject
        tasks = encoder.encode(src_dir, dest_dir, version=version, file_name=file_name, gui=gui, **kwargs)
        while tasks:  # многопоточно рекурсивно декодируем вложенные объекты MetaDataObject
            tasks_list = helper.run_in_pool(cls.encode_include, tasks, pool)
            tasks = helper.list_merge(*tasks_list)

    @classmethod
    def get_encoder(cls, src_dir, version='803'):
        if os.path.isfile(os.path.join(src_dir, 'ExternalDataProcessor.json')):
            versions = {
                '801': ExternalDataProcessor801,
                '802': ExternalDataProcessor802,
                '803': ExternalDataProcessor803,
            }
            return versions[version]
        if version == '803':
            if os.path.isfile(os.path.join(src_dir, 'ConfigurationExtension.json')):
                return ConfigurationExtension803
            if os.path.isfile(os.path.join(src_dir, 'Configuration.json')):
                return Configuration803
        raise FileNotFoundError(f'Не найдены файлы для сборки версии {version} ({src_dir})')

    @classmethod
    def encode_include(cls, include_type, encode_params):
        try:

            handler = helper.get_class_metadata_object(include_type)
            handler = handler.get_version(encode_params[3])
            tasks = handler.encode(*encode_params)
            return tasks
        except Exception as err:
            raise ExtException(parent=err, action=f'Decoder.encode_include({include_type})') from err


def encode(src_dir, dest_dir, *, pool=None, version='803', gui=None, file_name=None, **kwargs):
    Decoder.encode(src_dir, dest_dir, pool=pool, version=version, gui=gui, file_name=file_name, **kwargs)


def decode(src_dir, dest_dir, *, pool=None, version=None):
    Decoder.decode(src_dir, dest_dir, pool=pool, version=version)



import os
from base64 import b64decode, b64encode
from enum import Enum

from ..core.Simple import Simple
from ... import helper


class TmplType(Enum):
    table = "0"
    base64 = "1"
    html = "3"
    text = "4"
    scheme = "6"
    extension = "9"
    # todo добавить остальные типы макетов и их сериализацию


class Template8x(Simple):

    def __init__(self):
        super().__init__()
        self.data = None
        self.tmpl_type = None
        self.raw_data = None
        self.raw_header = None

    @property
    def uuid(self):
        if self.header:
            return self.header['uuid']
        return None

    def decode_object(self, src_dir, uuid, dest_dir, dest_path, version, header):
        super(Template8x, self).decode_object(src_dir, uuid, dest_dir, dest_path, version, header)
        self.tmpl_type = TmplType(self.get_template_type(header))
        self.header['type'] = self.tmpl_type.name

        _dest_dir = os.path.join(dest_dir, dest_path)
        try:
            getattr(self, f'decode_{self.tmpl_type.name}_data')(src_dir, _dest_dir, True)
        except AttributeError:
            raise Exception(f'Не реализованный тип макета {self.header["type"]}')

    @classmethod
    def get_decode_header(cls, header_data):
        return header_data[0][1][2]

    @classmethod
    def get_template_type(cls, header_data):
        return header_data[0][1][1]

    def decode_code(self, src_dir):
        return

    def decode_includes(self, src_dir, dest_dir, dest_path, header):
        return []

    def decode_scheme_data(self, src_dir, dest_dir, write):
        try:
            self.data = helper.bin_read(src_dir, f'{self.header["uuid"]}.0.bin')
            if write:
                helper.bin_write(self.data, dest_dir, f'{self.header["name"]}.bin')
        except FileNotFoundError:
            self.decode_text_data(src_dir, dest_dir, write)

    def decode_table_data(self, src_dir, dest_dir, write):
        self.decode_text_data(src_dir, dest_dir, write)

    def decode_text_data(self, src_dir, dest_dir, write):
        try:
            self.data, encoding = helper.txt_read_detect_encoding(src_dir, f'{self.header["uuid"]}.0.bin')
        except FileNotFoundError:
            return
        if write:
            helper.txt_write(self.data, dest_dir, f'{self.header["name"]}.bin', encoding=encoding)

    def decode_extension_data(self, src_dir, dest_dir, write):
        self.decode_base64_data(src_dir, dest_dir, write)

    def decode_base64_data(self, src_dir, dest_dir, write):
        try:
            data = helper.json_read(src_dir, f'{self.header["uuid"]}.0.json')
        except FileNotFoundError:
            return
        if data[0][1] and data[0][1][0]:
            self.data = b64decode(data[0][1][0][8:])
            data[0][1][0] = '"данные в отдельном файле"'
            extension = helper.get_extension_from_comment(self.header['comment'])
            if write:
                helper.bin_write(self.data, dest_dir, f'{self.header["name"]}.{extension}')

    def decode_html_data(self, src_dir, dest_dir, write):
        try:
            data = helper.json_read(src_dir, f'{self.header["uuid"]}.0.json')
        except FileNotFoundError:
            return
        if data[0][3] and data[0][3][0]:
            self.data = b64decode(data[0][3][0][8:])
            data[0][3][0] = '"данные в отдельном файле"'
            if write:
                helper.bin_write(self.data, dest_dir, f'{self.header["name"]}.html')

    def decode_header(self, header):
        self.tmpl_type = TmplType(header[0][1][1])
        self.header = {
            'type': self.tmpl_type.name,
        }
        _header = header[0][1][2]
        helper.decode_header(self.header, _header)

    def encode_object(self, src_dir, file_name, dest_dir, version):
        self.tmpl_type = TmplType[self.header['type']]

        self.raw_header = self.encode_header()
        try:
            handler = f'encode_{self.tmpl_type.name}_data'
            getattr(self, handler)(src_dir, dest_dir)
        except AttributeError:
            raise Exception(f'Не реализованный тип макета {self.header["type"]}')

        helper.json_write(self.raw_header, dest_dir, f'{self.header["uuid"]}.json')

    def encode_table_data(self, src_dir, dest_dir):
        self.encode_text_data(src_dir, dest_dir)

    def encode_scheme_data(self, src_dir, dest_dir):
        try:
            raw_data = helper.bin_read(src_dir, f'{self.header["name"]}.bin')
            helper.bin_write(raw_data, dest_dir, f'{self.header["uuid"]}.0.bin')
        except FileNotFoundError:
            self.encode_text_data(src_dir, dest_dir)

    def encode_text_data(self, src_dir, dest_dir):
        try:
            raw_data, encoding = helper.txt_read_detect_encoding(src_dir, f'{self.header["name"]}.bin')
            helper.txt_write(raw_data, dest_dir, f'{self.header["uuid"]}.0.bin', encoding=encoding)
        except FileNotFoundError:
            pass

    def encode_extension_data(self, src_dir, dest_dir):
        self.encode_base64_data(src_dir, dest_dir)

    def encode_base64_data(self, src_dir, dest_dir):
        extension = helper.get_extension_from_comment(self.header['comment'])
        bin_data = helper.bin_read(src_dir, f'{self.header["name"]}.{extension}')
        self._encode_bin_data(bin_data, dest_dir)

    def encode_html_data(self, src_dir, dest_dir):
        bin_data = helper.bin_read(src_dir, f'{self.header["name"]}.html')
        self.raw_data = [[
            "5",
            "1",
            "\"ru\"",
            ["#base64:" + b64encode(bin_data).decode(encoding='utf-8')],
            "0"
        ]]
        helper.json_write(self.raw_data, dest_dir, f'{self.header["uuid"]}.0.json')

    def _encode_bin_data(self, bin_data, dest_dir):
        self.raw_data = [[
            "1",
            ["#base64:" + b64encode(bin_data).decode(encoding='utf-8')]
        ]]
        helper.json_write(self.raw_data, dest_dir, f'{self.header["uuid"]}.0.json')

    def encode_header(self):
        raise NotImplemented()



import os
import shutil

from . import helper
from .code_organizer import CodeOrganizer
from .ext_exception import ExtException


class FileOrganizer:

    @classmethod
    def unpack(cls, src_dir, dest_dir, *, pool=None, index=None, descent=None):
        tasks = []
        cls._unpack(src_dir, os.path.abspath(dest_dir), '', tasks, index, descent)
        helper.run_in_pool(cls.unpack_code_file, tasks, pool=pool)

    @classmethod
    def _unpack(cls, src_dir, dest_dir, path, tasks, index, descent=None):
        entries = os.listdir(os.path.join(src_dir, path))
        for entry in entries:
            src_entry_path = os.path.join(src_dir, path, entry)

            if os.path.isdir(src_entry_path):
                new_path = os.path.join(path, entry)
                cls._unpack(src_dir, dest_dir, new_path, tasks, index, descent)
                continue
            if entry[-3:] == '.1c':
                tasks.append((src_dir, path, entry, dest_dir, index, descent))
            else:
                src_path = os.path.join(src_dir, path)
                cls.unpack_file(src_path, entry, dest_dir, path, entry, index, descent)

    @classmethod
    def unpack_file(cls, src_path, src_file_name, dest_dir, dest_path, dest_file_name, index, descent=None):
        dest_entry_path, dest_file_name = CodeOrganizer.get_dest_path(dest_dir, dest_path, dest_file_name, index,
                                                                      descent)
        dest_full_path = os.path.abspath(os.path.join(dest_dir, dest_entry_path))

        if dest_entry_path:
            os.makedirs(dest_full_path, exist_ok=True)

        src_full_path = os.path.join(src_path, src_file_name)

        if dest_full_path.startswith(dest_dir):  # файлы вне папки исходников не версионируются
            descent_full_dest_path, descent_file_name = cls.unpack_get_descent_filename(src_path, src_file_name, None,
                                                                                        dest_full_path, dest_file_name,
                                                                                        descent,
                                                                                        cls.equal_binary_file)
        else:
            descent_full_dest_path, descent_file_name = os.path.join(dest_dir, dest_entry_path), dest_file_name

        if descent_file_name:
            shutil.copy(src_full_path, os.path.join(descent_full_dest_path, descent_file_name))

    @classmethod
    def unpack_code_file(cls, src_dir, path, file_name, dest_dir, index, descent=None):
        try:
            code_areas = CodeOrganizer.unpack(src_dir, path, file_name, dest_dir, index)
            for elem in code_areas:
                _file = code_areas[elem]
                if elem == 'root':
                    _file['path'], _file['file_name'] = CodeOrganizer.get_dest_path(dest_dir, path, file_name, index,
                                                                                    descent)
                else:
                    _file['path'], _file['file_name'] = CodeOrganizer.parse_include_path(
                        elem, path, file_name, index.get('Области include') if index else None, descent)
                _file['dest_path'] = os.path.abspath(os.path.join(dest_dir, _file['path']))
                if _file['dest_path'].startswith(dest_dir):
                    descent_full_dest_path, descent_file_name = cls.unpack_get_descent_filename(
                        None, None, _file['data'], _file['dest_path'], _file['file_name'], descent,
                        cls.equal_code_file)
                else:
                    descent_full_dest_path, descent_file_name = _file['dest_path'], _file['file_name']

                if descent_file_name:
                    os.makedirs(descent_full_dest_path, exist_ok=True)
                    helper.txt_write(_file['data'], descent_full_dest_path, descent_file_name)
        except Exception as err:
            raise ExtException(
                parent=err,
                dump={"filename": os.path.join(src_dir, path, file_name)},
                action='unpack_code_file'
            )

    @classmethod
    def equal_binary_file(cls, src_path, src_file_name, src_data, near_descent_path, near_descent_file_name):
        with open(os.path.join(near_descent_path, near_descent_file_name), 'rb') as near_file:
            with open(os.path.join(src_path, src_file_name), 'rb') as src_file:
                near_data = near_file.read()
                src_data = src_file.read()
                if near_data == src_data:  # если файл не менялся, ничего с ним не делаем
                    return True
        return False

    @classmethod
    def equal_code_file(cls, src_path, src_file_name, src_data, near_descent_path, near_descent_file_name):
        near_data = helper.txt_read(near_descent_path, near_descent_file_name)
        if near_data == src_data:  # если файл не менялся, ничего с ним не делаем
            return True
        return False

    @classmethod
    def unpack_get_descent_filename(cls, src_path, src_file_name, src_data, dest_path, dest_file_name, descent,
                                    comparer):
        return dest_path, dest_file_name

    @classmethod
    def pack(cls, src_dir, dest_dir, *, pool=None, index=None, descent=None):
        helper.clear_dir(dest_dir)
        tasks = []
        src_dir = os.path.abspath(src_dir)
        index_code_areas = index.get('Области include') if index else None
        if index:
            cls._pack_index(src_dir, dest_dir, tasks, index, index_code_areas, [''], descent)
        cls._pack(src_dir, dest_dir, '', tasks, index, index_code_areas, descent)
        helper.run_in_pool(CodeOrganizer.pack, tasks, pool=pool)

    @classmethod
    def _pack_index(cls, src_dir: str, dest_dir: str, tasks: list, index: dict, index_code_areas: dict, path: list,
                    descent=None):
        for entry in index:
            if entry == 'Области include':
                continue
            if not index[entry]:
                continue
            if isinstance(index[entry], dict):
                path.append(entry)
                cls._pack_index(src_dir, dest_dir, tasks, index[entry], index_code_areas, path, descent)
                path.pop()
                pass
            elif isinstance(index[entry], str):
                if entry[-3:] == '.1c':
                    _src_path = os.path.join(
                        '..',
                        '' if descent is None else '..',  # в режиме с descent корень находится на уровень выше
                        os.path.dirname(index[entry])
                    )
                    _dest_path = os.path.join(*path)

                    _src_abs_path = os.path.abspath(_src_path)
                    if _src_abs_path.startswith(src_dir):
                        func_descent_filename = cls.pack_get_descent_filename
                    else:
                        func_descent_filename = FileOrganizer.pack_get_descent_filename
                    tasks.append((
                        src_dir, _src_path, os.path.basename(index[entry]),
                        dest_dir, _dest_path, entry, index_code_areas,
                        descent, func_descent_filename))
                else:
                    _dest_path = os.path.join(dest_dir, *path)
                    _src_full_path = os.path.join(
                        src_dir,
                        '..',
                        '' if descent is None else '..',  # в режиме с descent корень находится на уровень выше
                        index[entry]
                    )
                    _src_path = os.path.dirname(_src_full_path)
                    _src_file_name = os.path.basename(_src_full_path)
                    cls._pack_file(_src_path, _src_file_name, _dest_path, entry, descent)
            else:
                raise Exception('Некорректный формат файла индекса')

    @classmethod
    def pack_get_descent_filename(cls, src_path, src_file_name, descent):
        return src_path, src_file_name

    @classmethod
    def list_descent_dir(cls, src_dir, path, descent):
        return os.listdir(os.path.join(src_dir, path))

    @classmethod
    def _pack(cls, src_dir, dest_dir, path, tasks, index, index_code_areas, descent=None):
        if path:
            os.makedirs(os.path.join(dest_dir, path), exist_ok=True)
        entries = cls.list_descent_dir(src_dir, path, descent)
        for entry in entries:
            try:
                src_entry_path = os.path.join(src_dir, path, entry)

                if os.path.isdir(src_entry_path):
                    cls._pack(src_dir, dest_dir, os.path.join(path, entry), tasks, index, index_code_areas, descent)
                    continue
                _src_path = os.path.join(src_dir, path)
                descent_full_src_path, descent_file_name = cls.pack_get_descent_filename(
                    _src_path, entry, descent)

                if entry[-3:] == '.1c':
                    tasks.append((
                        src_dir, path, descent_file_name, dest_dir, path, entry, index_code_areas, descent,
                        cls.pack_get_descent_filename))
                else:
                    # shutil.copy(src_entry_path, os.path.join(dest_dir, path, entry))
                    _dest_path = os.path.join(dest_dir, path)
                    cls._pack_file(descent_full_src_path, descent_file_name, _dest_path, entry, descent)
            except Exception as err:
                raise ExtException(parent=err, action=f'{cls.__name__}._pack', detail=f'{entry} {path}')

    @classmethod
    def _pack_file(cls, src_path, src_file_name, dest_path, dest_file_name, descent=None):
        _src_path = os.path.join(src_path, src_file_name)
        _dest_path = os.path.join(dest_path, dest_file_name)
        try:
            shutil.copy(_src_path, _dest_path)
        except FileNotFoundError:
            _dest_dir = os.path.dirname(_dest_path)
            os.makedirs(_dest_dir, exist_ok=True)
            shutil.copy(_src_path, _dest_path)



import os

from .. import helper
from ..MetaObject import MetaObject
from ..ext_exception import ExtException


class MetaDataObject(MetaObject):
    versions = None
    version = None
    _obj_name = None

    def __init__(self):
        super().__init__()
        self.path = ''
        self.new_dest_path = None
        self.new_dest_dir = None
        self.new_dest_file_name = None

    @classmethod
    def get_obj_name(cls):
        return cls._obj_name if cls._obj_name else cls.__name__

    @classmethod
    def get_version(cls, version):
        if cls.versions is None:
            return cls
        try:
            return cls.versions[version]
        except KeyError:
            raise Exception(f'Нет реализации {cls.__name__} для версии "{version}"')

    @classmethod
    def decode(cls, src_dir, file_name, dest_dir, dest_path, version):
        try:
            header_data = helper.json_read(src_dir, f'{file_name}.json')
            self = cls()
            self.version = version
            self.decode_object(src_dir, file_name, dest_dir, dest_path, version, header_data)
            self.set_write_decode_mode(dest_dir, dest_path)
            tasks = self.decode_includes(src_dir, dest_dir, self.new_dest_path, header_data)
            self.write_decode_object(dest_dir, self.new_dest_path, self.new_dest_file_name, version)
            return tasks
        except ExtException as err:
            raise ExtException(
                parent=err,
                message=err.message + err.detail,
                detail=f'in decode {cls.__name__}',
                action=f'{cls.__name__}.decode'
            ) from err
        except Exception as err:
            raise ExtException(
                parent=err,
                action=f'{cls.__name__}.decode'
            ) from err

    def set_write_decode_mode(self, dest_dir, dest_path):
        self.set_mode_decode_in_root_folder(dest_dir, dest_path)

    def decode_object(self, src_dir, file_name, dest_dir, dest_path, version, header_data):
        self.set_header_data(header_data)

    def write_decode_object(self, dest_dir, dest_path, file_name, version):
        dest_full_path = os.path.join(dest_dir, dest_path)
        helper.json_write(self.header, dest_full_path, f'{file_name}.json')
        self.write_decode_code(dest_full_path, file_name)

    def set_mode_decode_in_name_folder(self, dest_dir, dest_path):
        self.new_dest_path = os.path.join(dest_path, self.header['name'])
        self.new_dest_dir = os.path.join(dest_dir, self.new_dest_path)
        self.new_dest_file_name = self.get_obj_name()
        os.makedirs(self.new_dest_dir)

    def set_mode_decode_in_root_folder(self, dest_dir, dest_path):
        self.new_dest_path = dest_path
        self.new_dest_dir = os.path.join(dest_dir, self.new_dest_path)
        self.new_dest_file_name = self.header['name']

    @classmethod
    def encode(cls, src_dir, file_name, dest_dir, version):
        self = cls()
        self.version = version
        file_name = self.get_encode_file_name(file_name)
        try:
            try:
                self.header = helper.json_read(src_dir, f'{file_name}.json')
            except FileNotFoundError:
                return
            self.encode_object(src_dir, file_name, dest_dir, version)
            self.write_encode_object(dest_dir)
            return self.encode_includes(src_dir, dest_dir)
        except Exception as err:
            raise ExtException(
                parent=err,
                dump=dict(src_dir=src_dir, file_name=file_name),
                action=f'{cls.__name__}.encode') from err

    @classmethod
    def get_encode_file_name(cls, file_name):
        return file_name

    def encode_object(self, src_dir, file_name, dest_dir, version):
        msg = f'Нет реализации для "{self.__class__.__name__}"'
        raise Exception(msg)

    def write_encode_object(self, dest_dir):
        helper.json_write(self.header['data'], dest_dir, f'{self.header["uuid"]}.json')
        self.write_encode_code(dest_dir)



from .ExternalDataProcessor803 import ExternalDataProcessor803


class ExternalDataProcessor802(ExternalDataProcessor803):
    version = '802'

    def encode_version(self):
        return [[
            [
                "216",
                "0"
            ]
        ]]



from .Simple import Simple
from .. import helper


class SimpleWithInfo(Simple):
    def decode_object(self, src_dir, file_name, dest_dir, dest_path, version, header_data):
        super(Simple, self).decode_object(src_dir, file_name, dest_dir, dest_path, version, header_data)
        try:
            self.header['info'] = helper.json_read(src_dir, f'{self.header["uuid"]}.0.json')
        except FileNotFoundError:
            return

    def write_encode_object(self, dest_dir):
        super(Simple, self).write_encode_object(dest_dir)
        info = self.header.get('info')
        if info:
            helper.json_write(info, dest_dir, f'{self.header["uuid"]}.0.json')



# -*- coding: utf-8 -*-
import datetime
import io
import os
import tempfile
import zlib
from base64 import b64encode
from hashlib import sha1
from struct import pack, calcsize

from . import helper
from .json_container_decoder import JsonContainerDecoder

# INT32_MAX
END_MARKER = 2147483647
DEFAULT_BLOCK_SIZE = 512  # 0x200
# Для формата старше 8.3.15
DEFAULT_BLOCK_SIZE64 = 65536  # 0x1000
# Размер буффера передачи данных из потока в поток
BUFFER_CHUNK_SIZE = 512


def epoch2int(epoch_time):
    """
    Преобразует время в формате "количество секунд с начала эпохи" в количество сотых микросекундных интервалов
    с 0001.01.01

    :param epoch_time: время в формате Python
    :type epoch_time: real
    :return: количество сотых микросекундных интервалов
    :rtype: int
    """
    # Начало эпохи на разных системах - разная дата
    # Поэтому явно вычисляем разницу между указанной датой и 0001.01.01
    return (datetime.datetime.fromtimestamp(epoch_time) - datetime.datetime(1, 1, 1)) // datetime.timedelta(
        microseconds=100)


def int2hex(value):
    """
    Получает строковое представление целого числа в шестнадцатиричном формате длиной не менее 4 байт

    :param value: конвертируемое число
    :type value: int
    :return: предоставление числа
    :type: string
    """
    return '{:02x}'.format(value).rjust(8, '0')


def get_size(file):
    """
    Возвращает размер file-like объекта

    :param file: объекта файла
    :type file: BufferedReader
    :return: размер в байтах
    :rtype: int
    """
    pos = file.tell()
    file.seek(0, os.SEEK_END)
    size = file.tell()
    file.seek(pos)
    return size


class ContainerWriter(object):
    """
    Класс для записи контейнеров

    :param file: объект файла контейнера
    :type file: BufferedReader
    """

    def __init__(self, file):
        self.file = file
        self.toc = []

    def write_header(self):
        """
        Записывает заголовок контейнера

        """
        self.file.write(pack('4i', END_MARKER, DEFAULT_BLOCK_SIZE, 0, 0))

    def write_block(self, data, **kwargs):
        """
        Записывает блок данных в контейнер

        :param data: file-like объект
        :param kwargs: Опциональные параметры
        :return: смещение записанных данных (байт)
        :rtype: int
        """
        # размер данных блока
        size = kwargs.pop('size', get_size(data))
        offset = kwargs.pop('offset', get_size(self.file))
        self.file.seek(offset)

        block_size = kwargs.pop('block_size', max(DEFAULT_BLOCK_SIZE, size))
        next_block_offset = kwargs.pop('next_block_offset', END_MARKER)

        if len(kwargs) > 0:
            raise ValueError('Unsupported arguments: {}'.format(','.join(kwargs.keys())))

        header_data = ('\r\n', int2hex(size), ' ', int2hex(block_size), ' ', int2hex(next_block_offset), ' \r\n')
        header = pack('2s8ss8ss8s3s', *[x.encode() for x in header_data])

        self.file.write(header)
        self.write_block_data(self, data, self.file)

        self.file.write(b'\x00' * (block_size - data.tell()))

        return offset

    @staticmethod
    def write_block_data(container, data, dest_file):
        data.seek(0)
        while True:
            buffer = data.read(BUFFER_CHUNK_SIZE)
            if not buffer:
                break
            dest_file.write(buffer)

    def add_file(self, fd, name, inflate=False):
        """
        Добавляет файл в контейнер

        :param fd: file-like объект файла
        :type fd: BufferedReader
        :param name: Имя файла в контейнере
        :type name: string
        :param inflate: флаг сжатия
        :type inflate: bool
        """
        modify_time = epoch2int(os.stat(fd.fileno()).st_mtime)
        # В *nix это не время создания файла.
        creation_time = epoch2int(os.stat(fd.fileno()).st_ctime)

        buffer = b''.join([pack('QQi', creation_time, modify_time, 0), name.encode('utf-16-le'), b'\x00' * 4])
        attribute_doc_offset = self.write_block(io.BytesIO(buffer), block_size=len(buffer))

        if inflate:
            data_doc_offset = self.compress(self, fd, self.file, self.write_block)
        else:
            data_doc_offset = self.write_block(fd)

        self.toc.append((attribute_doc_offset, data_doc_offset))

    @staticmethod
    def compress(container, src_fd, dest_fd, block_writer):
        with tempfile.TemporaryFile() as f:
            compressor = zlib.compressobj(wbits=-15)
            src_fd.seek(0)
            while True:
                chunk = src_fd.read(BUFFER_CHUNK_SIZE)
                if not chunk:
                    f.write(compressor.flush())
                    break
                f.write(compressor.compress(chunk))
            data_doc_offset = block_writer(container, f, dest_fd)
        return data_doc_offset

    def write_toc(self):
        """
        Записывает оглавление контейнера
        """
        if len(self.toc) == 0:
            raise IOError('Container is empty')
        with tempfile.TemporaryFile() as f:
            for attr_offset, data_offset in self.toc:
                f.write(b''.join([pack('3i', attr_offset, data_offset, END_MARKER)]))

            size = get_size(f)
            total_blocks = size // DEFAULT_BLOCK_SIZE + 1

            if total_blocks == 1:
                self.write_block(f, size=size, offset=calcsize('4i'))
            else:
                f.seek(0)
                next_block_offset = get_size(self.file)
                self.write_block(io.BytesIO(f.read(DEFAULT_BLOCK_SIZE)), size=size, offset=calcsize('4i'),
                                 next_block_offset=next_block_offset, block_size=DEFAULT_BLOCK_SIZE)
                for i in range(1, total_blocks):
                    # 31 - длина заголовка блока
                    next_block_offset += DEFAULT_BLOCK_SIZE + 31
                    self.write_block(io.BytesIO(f.read(DEFAULT_BLOCK_SIZE)), size=0,
                                     next_block_offset=next_block_offset)
                self.write_block(io.BytesIO(f.read(DEFAULT_BLOCK_SIZE)), size=0)

    def __enter__(self):
        """
        Вход в блок. Позволяет применять оператор with.
        """
        self.write_header()
        self.file.write(b'\x00' * (DEFAULT_BLOCK_SIZE + 31))
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Выход из блока. Позволяет применять оператор with.
        """
        self.write_toc()


def add_entries(container, folder, nested=False):
    """
    Рекурсивно добавляет файлы из директории в контейнер

    :param container: объект файла контейнера
    :type container: BufferedReader
    :param folder: каталог файлов, которые надо поместить в контейнер
    :type folder: string
    :param nested: обрабатывать вложенные каталоги
    :type nested: bool
    """
    entries = sorted(os.listdir(folder))
    for entry in entries:
        entry_path = os.path.join(folder, entry)
        if os.path.isdir(entry_path):
            with tempfile.TemporaryFile() as tmp:
                with ContainerWriter(tmp) as nested_container:
                    add_entries(nested_container, entry_path, nested=True)
                container.add_file(tmp, entry, inflate=not nested)
        else:
            with open(entry_path, 'rb') as entry_file:
                container.add_file(entry_file, entry, inflate=not nested)


def build(folder, filename, nested=False):
    """
    Запаковывает каталог в контейнер включая вложенные каталоги.
    Сахар для ContainerWriter.

    :param folder: каталог с данными, запаковываемыми в контейнер
    :type folder: string
    :param filename: имя файла контейнера
    :type filename: string
    """
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    with open(filename, 'w+b') as f, ContainerWriter(f) as container:
        add_entries(container, folder, nested)


def calc_sha1(src_folder, dest_folder):
    entries = sorted(os.listdir(dest_folder))
    versions = []
    versions_file = ''
    for filename in entries:
        dest_path = os.path.join(dest_folder, filename)
        if filename == 'configinfo':
            versions_file = 'configinfo'
            continue

        with open(dest_path, 'rb') as file:
            data = file.read()
        versions.append(f'"{filename}"')
        versions.append(b64encode(sha1(data).digest()).decode())
    if not versions_file:
        return
    versions.insert(0, str(int(len(versions) / 2)))
    _versions = JsonContainerDecoder().encode_root_object([versions])

    src_path = os.path.join(src_folder, versions_file)
    dest_path = os.path.join(dest_folder, versions_file)

    with open(src_path, 'r+', encoding='utf-8') as file:
        data = file.read()
        data = data.replace('{____versions____}', _versions)
        file.seek(0)
        file.write(data)

    compress_and_build_simple_file(src_path, dest_path)
    pass


def compress_and_build_simple_file(src_path, dest_path):
    with open(dest_path, 'w+b') as dest_fd:
        with open(src_path, 'rb') as src_fd:
            ContainerWriter.compress(None, src_fd, dest_fd, ContainerWriter.write_block_data)


def compress_and_build(src_folder, dest_folder, *, pool=None, nested=False):
    helper.clear_dir(dest_folder)
    entries = sorted(os.listdir(src_folder))

    for filename in entries:
        src_path = os.path.join(src_folder, filename)
        dest_path = os.path.join(dest_folder, filename)
        # add_entry(container, src_folder, filename)
        if os.path.isdir(src_path):
            with open(dest_path, 'w+b') as dest_fd:
                with tempfile.TemporaryFile() as tmp:
                    with ContainerWriter(tmp) as nested_container:
                        add_entries(nested_container, src_path, nested=True)
                    ContainerWriter.compress(None, tmp, dest_fd, ContainerWriter.write_block_data)
        else:
            compress_and_build_simple_file(src_path, dest_path)
    calc_sha1(src_folder, dest_folder)



import json
import os
import shutil
from codecs import BOM_UTF8, BOM_UTF16_BE, BOM_UTF16_LE, BOM_UTF32_BE, BOM_UTF32_LE
from multiprocessing import Pool, cpu_count

from .ext_exception import ExtException


def json_read(path, file_name):
    _path = os.path.join(path, file_name)
    with open(_path, 'r', encoding='utf-8') as file:
        return json.load(file)


def json_write(data, path, file_name):
    _path = os.path.join(path, file_name)
    os.makedirs(path, exist_ok=True)
    with open(_path, 'w', encoding='utf-8') as file:
        json.dump(data, file, ensure_ascii=False, indent=2)


def txt_read(path, file_name, encoding='utf-8-sig'):
    return txt_read_detect_encoding(path, file_name, encoding=encoding)[0]


def txt_read_detect_encoding(path, file_name, encoding='utf-8'):
    _path = os.path.join(path, file_name)
    if encoding is None:
        encoding = detect_by_bom(_path, 'utf-8')
    with open(_path, 'r', encoding=encoding) as file:
        return file.read(), encoding


def txt_write(data, path, file_name, encoding='utf-8'):
    if data is None:
        return
    _path = os.path.join(path, file_name)
    os.makedirs(path, exist_ok=True)
    with open(_path, 'w', encoding=encoding) as file:
        file.write(data)


def bin_write(data, path, file_name):
    _path = os.path.join(path, file_name)
    os.makedirs(path, exist_ok=True)
    with open(_path, 'wb') as file:
        file.write(data)


def bin_read(path, file_name):
    _path = os.path.join(path, file_name)
    with open(_path, 'rb') as file:
        return file.read()


def decode_header(obj: dict, header: list):
    obj['uuid'] = header[1][2]
    obj['name'] = str_decode(header[2])
    obj['name2'] = {}
    count_locale = int(header[3][0])
    for i in range(count_locale):
        obj['name2'][str_decode(header[3][i * 2 + 1])] = str_decode(header[3][i * 2 + 2])
    comment = str_decode(header[4]).split(';')[0]  # удаляем имя файла и номер версии которую добавляем при сборке
    header[4] = str_encode(comment)
    obj['comment'] = comment
    obj['h1_0'] = header[1][0]
    obj['h0'] = header[0]
    obj['h5'] = header[5:]


def clear_dir(path: str) -> None:
    shutil.rmtree(path, ignore_errors=True)
    os.makedirs(path, exist_ok=True)


def str_encode(data: str) -> str:
    return f'"{data}"'


def encode_name2(header: dict):
    result = [str(len(header['name2']))]
    for elem in header['name2']:
        result.append(str_encode(elem))
        result.append(str_encode(header['name2'][elem]))
    return result


def str_decode(data: str) -> str:
    return data[1:-1]


def get_pool(*, pool: Pool = None, processes=None) -> Pool:
    if pool is not None:
        return pool
    if processes is None:
        processes = cpu_count()
    return Pool(processes=processes)


def close_pool(local_pool: Pool, pool: Pool = None) -> None:
    if pool is None:
        local_pool.close()
        local_pool.join()


def run_in_pool(method, list_args, pool=None):
    _pool = get_pool(pool=pool)
    # msg = f'pool {method}({len(list_args)})'
    try:
        result = _pool.starmap(method, list_args)
    except ExtException as err:
        raise ExtException(
            parent=err,
            action=f'run_in_pool {method.__qualname__}') from err
    # except Exception as err:
    #     raise ExtException(parent=err, detail=f'{method.__qualname__ {err.message}' action=f'run_in_pool {method.__qualname__}') from err
    finally:
        close_pool(_pool, pool)
    return result


def list_merge(*args):
    result = []
    for lst in args:
        if lst:
            result.extend(lst)
    return result


def get_class_metadata_object(name):
    return get_class(f'v8unpack.MetaDataObject.{name}.{name}')


def get_class(kls):
    try:
        parts = kls.split('.')
        module = ".".join(parts[:-1])
        m = __import__(module)
        for comp in parts[1:]:
            m = getattr(m, comp)
        return m
    except ImportError as e:
        # ошибки в классе  или нет файла
        raise ImportError(f'get_class({kls}: {str(e)}')
    except AttributeError as e:
        # Нет такого класса
        raise AttributeError(f'get_class({kls}: {str(e)}')
    except Exception as e:
        # ошибки в классе
        raise Exception(f'get_class({kls}: {str(e)}')


def detect_by_bom(path, default=None):
    boms = (
        ('utf-8-sig', BOM_UTF8),
        ('utf-32', BOM_UTF32_LE),
        ('utf-32', BOM_UTF32_BE),
        ('utf-16', BOM_UTF16_LE),
        ('utf-16', BOM_UTF16_BE),
    )

    with open(path, 'rb') as f:
        raw = f.read(4)  # will read less if the file is smaller
    for enc, bom in boms:
        if raw.startswith(bom):
            return enc
    return default


def str_time(value, _format='%H:%M:%S.%f'):
    return value.strftime(_format)


def get_extension_from_comment(comment: str) -> str:
    comment = comment.strip()
    res = 'bin'
    if comment:
        ext = comment.split(" ")[-1]
        if len(ext) < 6 and ext.isalnum():
            return ext
    return res


def check_version(v8_version: str, src_version: str) -> None:
    _v8 = v8_version.split(".")
    _src = src_version.split(".")
    if len(_v8) != 3:
        raise AssertionError(f'Не правильный номер версии v8unpack "{v8_version}"')
    if len(_src) != 3:
        raise AssertionError(f'Не правильная версия исходников "{src_version}"')
    if (int(_v8[0]), int(_v8[1])) != (int(_src[0]), int(_src[1])):
        raise AssertionError(
            f"Версия исходников {_src[0]}.{_src[1]} не соответствует версии v8unpack {_v8[0]}.{_v8[1]}")


def get_descent_file_name(file_name, descent):
    name: list = file_name.split('.')
    name.insert(-1, str(descent))
    return '.'.join(name)


def get_near_descent_file_name(path, file_name, descent):
    name: list = file_name.split('.')
    startswith = '.'.join(name[0:-1])
    endswith = name[-1]
    size = len(name)
    try:
        entities = os.listdir(path)
    except FileNotFoundError:
        os.makedirs(path, exist_ok=True)
        entities = []
    descents = []
    for entity in entities:
        if entity.startswith(startswith) and entity.endswith(endswith):
            _entity = entity.split('.')
            if len(_entity) - 1 != size:
                continue
            full_path = os.path.join(path, entity)
            if os.path.isfile(full_path):
                try:
                    descents.append(int(_entity[-2]))
                except ValueError:  # если во втором разряде не число, значит не наш вариант
                    pass
    if not descents:
        return '', ''
    descents = sorted(descents, reverse=True, key=lambda x: 0 if x > descent else x)
    if descents[0] > descent:
        return '', ''
    return path, f'{startswith}.{descents[0]}.{endswith}'


def remove_descent_from_filename(file_name):
    _name = file_name.split('.')
    try:
        if len(_name) < 3:
            return file_name
        _descent = _name.pop(-2)
        if str(int(_descent)) == _descent:
            return '.'.join(_name)
    except Exception:
        return file_name



import os

from ..core.Simple import SimpleNameFolder
from ... import helper

FORM83 = "13"
FORM82 = "9"
FORM81 = "7"


class Form8x(SimpleNameFolder):
    ver = ''
    _obj_name = "Form"

    def __init__(self):
        super().__init__()
        self.form = []
        self.elements = None

    def decode_object(self, src_dir, uuid, dest_dir, dest_path, version, header_data):
        self.set_header_data(header_data)
        self.decode_data(src_dir, uuid)
        self.decode_code(src_dir)

    def decode_old_form(self, src_dir):
        _code_dir = f'{os.path.join(src_dir, self.header["uuid"])}.0'
        if os.path.isdir(_code_dir):
            self.form.append(helper.json_read(_code_dir, 'form.json'))
            try:
                encoding = helper.detect_by_bom(os.path.join(_code_dir, 'module.bin'), 'utf-8')
                self.code['obj'] = self.read_raw_code(_code_dir, 'module.bin', encoding=encoding)
            except UnicodeDecodeError:
                encoding = 'windows-1251'
                self.code['obj'] = self.read_raw_code(_code_dir, 'module.bin', encoding=encoding)
            self.header['code_encoding_obj'] = encoding  # можно безболезненно поменять на utf-8-sig
            self.header[f'code_info_obj'] = 1

    def write_decode_object(self, dest_dir, dest_path, file_name, version):
        super(Form8x, self).write_decode_object(dest_dir, dest_path, file_name, version)
        helper.json_write(self.form, self.new_dest_dir, f'{file_name}.form{self.ver}.json')
        return []

    def decode_data(self, src_dir, uuid):
        raise NotImplemented()

    @classmethod
    def get_decode_header(cls, header):
        return cls.get_decode_obj_header(header)[1][1]

    @classmethod
    def get_decode_obj_header(cls, header):
        return header[0][1][1]

    def encode_object(self, src_dir, file_name, dest_dir, version):
        super(Form8x, self).encode_object(src_dir, file_name, dest_dir, version)
        try:
            self.form = helper.json_read(src_dir, f'{file_name}.form{version}.json')
        except FileNotFoundError:
            self.form = self.encode_empty_form()
        self.encode_data()

    def encode_header(self):
        raise NotImplemented()

    def encode_data(self):
        raise NotImplemented()

    def encode_empty_form(self):
        raise NotImplemented()

    def encode_write(self, dest_dir):
        raise NotImplemented()

    def write_encode_code(self, dest_dir):
        pass

    def write_old_encode_object(self, dest_dir):
        helper.json_write(self.encode_header(), dest_dir, f'{self.header["uuid"]}.json')
        if self.header.get('code_info_obj'):
            _code_dir = f'{os.path.join(dest_dir, self.header["uuid"])}.0'
            os.makedirs(_code_dir, exist_ok=True)
            helper.json_write(self.form[0], _code_dir, 'form.json')
            _encoding = self.header.get('code_encoding_obj', 'utf-8-sig')
            self.write_raw_code(self.code.get('obj', ''), _code_dir, 'module.bin', encoding=_encoding)



import json
import sys
import traceback
from os import path


# 3xxx - Ошибки в нашем коде
# 4xxx - Ошибки вызванные некорректными действиями пользователей
# 5xxx - Ошибки вызванные временной недоступностью функционала, повторный вызов с теми же параметрами может быть успешным
# 6xxx - Ошибки во внешних сервисах и библиотеках
# 8xxx - Запросы не реализованного, но потенциально интересного функционала
# 9xxx - Обертка над стандартными ошибками для использования в коде


class ExtException(Exception):
    _code = 3000
    _http_code = 500
    _message = "Unknown error"

    def __new__(cls, *args, **kwargs):
        parent = kwargs.get('parent')
        if parent:
            # прокидываем ошибку наверх
            if isinstance(parent, ExtException) and parent.__class__ != ExtException and cls == ExtException:
                return parent.__class__(*args, **kwargs)
            # восстанавливаем из строки
        return super(ExtException, cls).__new__(cls, *args, **kwargs)

    def __init__(self, *args, **kwargs):
        self.skip_traceback = kwargs.get('skip_traceback', 0)
        self.code = ''
        self.message = ''
        self.detail = ''
        self.action = ''
        self.dump = {}
        self.stack = []
        self.new_msg = False
        parent = kwargs.get('parent')
        if parent and isinstance(parent, ExtException):  # прокидываем ошибку наверх
            self.add_parent_to_stack(parent)
            self.message = parent.message
            self.code = parent.code
            self.detail = parent.detail
            self.new_msg = bool(kwargs.get('message'))
        self.init_from_dict(kwargs)
        if parent and isinstance(parent, Exception) and not isinstance(parent, ExtException):
            self.code = self.code if self.code else self._code
            if not self.detail:
                self.add_sys_exc_to_stack()
                self.detail = f'{parent.__class__.__name__} {str(parent)}'
        if args:
            if isinstance(args[0], str):
                self.new_msg = True
                self.message = args[0]
            else:
                raise Exception(f'Not supported ExtException args[0] ({type(args[0])})')
        if not self.stack:
            self.add_sys_exc_to_stack()
        pass

    def add_parent_to_stack(self, parent):
        if not isinstance(parent, ExtException):
            return None
        self.stack += parent.stack
        if not parent.action:
            return
        parent.dump['action'] = parent.action
        if not self.stack:
            data = self.get_sys_exc_info()
            if data:
                parent.dump['traceback'] = data['traceback']

        if parent.new_msg:
            parent.dump['message'] = parent.message
            if parent.detail:
                parent.dump['detail'] = parent.detail
        self.stack.append(parent.dump)

    def add_sys_exc_to_stack(self):
        try:
            data = self.get_sys_exc_info()
        except IndexError:
            return
        if not data:
            return
        data['action'] = self.action
        self.stack.append(data)

    def get_sys_exc_info(self):
        exc_info = sys.exc_info()
        if exc_info[2] is None:
            return None
        last_call = traceback.extract_tb(exc_info[2], limit=self.skip_traceback + 1)
        last_call = last_call[self.skip_traceback]
        return {
            'message': exc_info[0].__name__,
            'detail': str(exc_info[1]),
            'traceback': f'{path.basename(path.dirname(last_call.filename))}/{path.basename(last_call.filename)}, '
                         f'{last_call.name}, line {last_call.lineno}'
        }

    def init_from_dict(self, data):
        for field in ['code', 'message', 'detail', 'action', 'dump', 'stack']:
            if field in data:
                setattr(self, field, data[field])
        if not self.message:
            self.message = self._message
        if not self.code:
            self.code = self._code

    @property
    def title(self):
        if self.detail:
            return f'{self.code}: {self.message} - {self.detail} {self.action}'
        return f'{self.code}: {self.message}'

    def __str__(self):
        res = f'{self.title}'
        if self.dump:
            res += f'\nDump: action={self.action}; '
            for elem in self.dump:
                res += f'{elem}={self.dump[elem]}; '
        if self.stack:
            res += '\n Stack:'
            for stack in self.stack:
                _action = stack.get('action', '')
                res += f'\n  - {_action}: '
                for elem in stack:
                    if elem != 'action':
                        res += f'{elem}={stack[elem]}; '
        return res

    def dumps(self):
        return json.dumps(self.to_dict(), ensure_ascii=False)

    def to_dict(self):
        return {
            '__name__': self.__class__.__name__,
            '__module__': self.__class__.__module__,
            'message': self.message,
            'detail': self.detail,
            'code': self.code,
            'action': self.action,
            'dump': self.dump,
            'stack': self.stack
        }

    @property
    def http_code(self):
        return self._http_code


class HandlerNotFoundError(ExtException):
    _code = 3100
    _message = "Handler not found",


class UserError(ExtException):
    _code = 4000


class Unauthorized(ExtException):
    _code = 4011
    _http_code = 401
    _message = "Unauthorized"


class ResourceNotAvailable(ExtException):
    _code = 4100
    _message = "Resource not available"


class KeyNotFound(ExtException):
    _code = 4200
    _message = "Required parameter are missing"


class AccessDenied(ExtException):
    _code = 4031
    _http_code = 403
    _message = "Access denied"


class CancelOperation(ExtException):
    _code = 4999
    _message = "Cancel operation",


class NotAvailableError(ExtException):
    _code = 5000
    _message = "Not available",


class ForeignError(ExtException):
    _code = 6000
    _message = "Unknown external error",


class ExtTimeoutError(ExtException):
    _code = 5008
    _http_code = 408
    _message = "Timeout",


def dumps_error(err):
    return json.dumps({
        '__name__': err.__class__.__name__,
        '__module__': err.__class__.__module__,
        'message': str(err)
    })



import os

from . import helper
from .file_organizer import FileOrganizer


class FileOrganizerCE(FileOrganizer):

    @classmethod
    def unpack_get_descent_filename(cls, src_path, src_file_name, src_data, dest_path, dest_file_name, descent,
                                    comparer):
        # если есть файл нашей версии или новый файл отличается от старого, или старого файла нет
        # сохраняем файл с номером текущей версии
        # считаем, что общие изменения должны делаться на минимальной версии
        descent_file_name = helper.get_descent_file_name(dest_file_name, descent)
        descent_full_dest_path = os.path.join(dest_path, descent_file_name)

        if not os.path.isfile(descent_full_dest_path):  # если нет файла нужной версии
            near_descent_path, near_descent_file_name = helper.get_near_descent_file_name(dest_path, dest_file_name,
                                                                                          descent)
            if near_descent_file_name:  # нашли файл младшей версии
                if comparer(src_path, src_file_name, src_data, near_descent_path, near_descent_file_name):
                    return '', ''
        return dest_path, descent_file_name

    @classmethod
    def pack_get_descent_filename(cls, src_path, src_file_name, descent):
        descent_path, descent_file_name = helper.get_near_descent_file_name(src_path, src_file_name, descent)
        if not descent_file_name:
            raise FileNotFoundError(f'{src_path}/{src_file_name} ({descent})')
        return descent_path, descent_file_name

    @classmethod
    def list_descent_dir(cls, src_dir, path, descent):
        def check_descent_name(_name):
            try:
                if len(_name) < 3:
                    return False
                _descent = _name[-2]
                if str(int(_descent)) == _descent:
                    return True
            except Exception:
                return False

        _index = {}
        result = []
        _dir = os.path.join(src_dir, path)
        entries = os.listdir(_dir)
        for entry in entries:
            full_path = os.path.join(_dir, entry)
            if os.path.isdir(full_path):
                result.append(entry)
            else:
                entry = helper.remove_descent_from_filename(entry)
                if entry not in _index:
                    _index[entry] = 1
                    result.append(entry)
        return result



from .. import __version__
from .. import helper
from ..MetaObject.Configuration803 import Configuration803


class ConfigurationExtension803(Configuration803):
    info = ['6', '8', '9']
    ext_code = {
        'con': '5',
        'app': '6',
        'ssn': '7'
    }

    def __init__(self):
        super(ConfigurationExtension803, self).__init__()

    @classmethod
    def decode(cls, src_dir, dest_dir, *, version=None):
        self = cls()
        self.header = {}
        root = helper.json_read(src_dir, 'configinfo.json')
        self.header['v8unpack'] = __version__
        self.header['file_uuid'] = root[1][1]
        self.header['version'] = root[0][1]
        # self.header['versions'] = root[2]
        self.header['copyinfo'] = root[1]
        # self.header['copyinfo'][2] = b64decode(self.header['copyinfo'][2])[:-16].hex()
        self.header['data'] = helper.json_read(src_dir, f'{self.header["file_uuid"]}.json')
        _form_header = self.get_decode_header(self.header['data'])
        helper.decode_header(self.header, _form_header)
        product_version = self.header['data'][0][3][1][1][15]
        if version is None:
            self.header['compatibility_version'] = self.header['data'][0][3][1][1][43]
        else:
            self.header['compatibility_version'] = version
            self.header['data'][0][3][1][1][43] = version

        self.decode_code(src_dir)

        for i in self.info:  # хз что это
            try:
                self.header[f'info{i}'] = helper.json_read(src_dir, f'{self.header["uuid"]}.{i}.json')
            except FileNotFoundError:
                pass
        helper.txt_write(helper.str_decode(product_version), dest_dir, 'version.bin', encoding='utf-8')
        helper.json_write(self.header, dest_dir, f'{cls.get_class_name_without_version()}.json')
        self.write_decode_code(dest_dir, cls.__name__)
        tasks = self.decode_includes(src_dir, dest_dir, '', self.header['data'])
        return tasks

    @classmethod
    def encode(cls, src_dir, dest_dir, *, version=None, file_name=None, gui=None, **kwargs):
        self = cls()
        helper.clear_dir(dest_dir)
        self.header = helper.json_read(src_dir, f'{cls.get_class_name_without_version()}.json')

        self.set_product_info(src_dir, file_name)

        if version is not None:
            self.header['data'][0][3][1][1][43] = version
        if gui is not None:
            self.header['data'][0][3][1][1][38] = gui

        helper.check_version(__version__, self.header.get('v8unpack', ''))

        # self.header['copyinfo'][2] = b64encode(bytes.fromhex(self.header['copyinfo'][2]+md5().digest().hex())).decode()
        root = [
            ["0", self.encode_version()],
            self.header['copyinfo'],
            # self.header['versions']
            ["____versions____"]
        ]
        self.encode_code(src_dir, cls.__name__)
        self.write_encode_code(dest_dir)
        helper.json_write(root, dest_dir, 'configinfo.json')
        helper.json_write(self.header['data'], dest_dir, f'{self.header["file_uuid"]}.json')
        for i in self.info:  # хз что это
            try:
                helper.json_write(self.header[f'info{i}'], dest_dir, f'{self.header["uuid"]}.{i}.json')
            except KeyError:
                pass
        tasks = self.encode_includes(src_dir, dest_dir)
        return tasks

    def set_product_version(self, product_version):
        self.header['data'][0][3][1][1][15] = helper.str_encode(product_version)



from .. import __version__
from .. import helper
from ..MetaObject import MetaObject


class Configuration803(MetaObject):

    def __init__(self):
        super(Configuration803, self).__init__()
        self.counter = {}

    @classmethod
    def decode(cls, src_dir, dest_dir, *, version=None):
        self = cls()
        self.header = {}
        root = helper.json_read(src_dir, 'root.json')
        self.header['v8unpack'] = __version__
        self.header["file_uuid"] = root[0][1]
        self.header["root_data"] = root

        _header_data = helper.json_read(src_dir, f'{self.header["file_uuid"]}.json')
        self.set_header_data(_header_data)

        self.header['version'] = helper.json_read(src_dir, 'version.json')
        self.header['versions'] = helper.json_read(src_dir, 'versions.json')

        helper.json_write(self.header, dest_dir, f'{cls.get_class_name_without_version()}.json')

        tasks = self.decode_includes(src_dir, dest_dir, '', self.header['data'])
        return tasks
        pass

    @classmethod
    def get_decode_header(cls, header):
        return header[0][3][1][1][1][1]

    @classmethod
    def get_decode_includes(cls, header_data):
        return [
            header_data[0][3][1],
            header_data[0][4][1][1],
            header_data[0][5][1],
            header_data[0][6][1],
            header_data[0][7][1],
            header_data[0][8][1],
        ]

    @classmethod
    def encode(cls, src_dir, dest_dir, *, version=None, file_name=None, **kwargs):
        self = cls()
        helper.clear_dir(dest_dir)
        self.header = helper.json_read(src_dir, f'{cls.get_class_name_without_version()}.json')
        helper.check_version(__version__, self.header.get('v8unpack', ''))

        helper.json_write(self.header["root_data"], dest_dir, 'root.json')
        helper.json_write(self.encode_version(), dest_dir, 'version.json')
        helper.json_write(self.header["versions"], dest_dir, 'versions.json')

        helper.json_write(self.header['data'], dest_dir, f'{self.header["file_uuid"]}.json')
        tasks = self.encode_includes(src_dir, dest_dir)
        return tasks

    def encode_version(self):
        return self.header['version']
